#include <stdio.h>
#include "par_for.h"
//for_loop runs n iterations applying body(i,arg_body)
void for_loop(int i, int n, void (*body)(int, void *), void *arg_body){
 if (n==1){
  body(i, arg_body);
  return;
 }
 if (n==2){
  body(i, arg_body); body(i+1, arg_body);
  return;
 }
 for_loop(i, n/2, body, arg_body);
 for_loop(i+n/2, n-n/2, body, arg_body);
}
//for_ex_cond runs n iterations applying body(i,arg_body) if cond(i,arg_cond) is false
//cond is an excluding condition : iteration i is empty when cond(i,arg_cond) is true
//for_ex_cond returns the number of included iterations (from 0 to n)
int for_ex_cond(int i, int n, int (*cond)(int, void *), void *arg_cond, void (*body)(int, void *), void *arg_body){
 int c1, c2;
 if (n==1){
  c1=cond(i,arg_cond); if (!c1) body(i,arg_body);
  return !c1;
 }
 if (n==2){
  c1=cond(i,arg_cond); if (!c1) body(i,arg_body);
  c2=cond(i+1,arg_cond); if (!c2) body(i+1,arg_body);
  return (!c1) + (!c2);
 }
 c1=for_ex_cond(i,n/2,cond,arg_cond,body,arg_body);
 c2=for_ex_cond(i+n/2,n-n/2,cond,arg_cond,body,arg_body);
 return c1+c2;
}
//for_in_cond runs n iterations applying body(i,arg_body) if cond(i,arg_cond) is true
//cond is an including condition : iteration i is empty when cond(i,arg_cond) is false
//for_in_cond returns the number of excluded iterations (from 0 to n)
int for_in_cond(int i, int n, int (*cond)(int, void *), void *arg_cond, void (*body)(int, void *), void *arg_body){
 int c1, c2;
 if (n==1){
  c1=cond(i,arg_cond); if (c1) body(i,arg_body);
  return !c1;
 }
 if (n==2){
  c1=cond(i,arg_cond); if (c1) body(i,arg_body);
  c2=cond(i+1,arg_cond); if (c2) body(i+1,arg_body);
  return (!c1) + (!c2);
 }
 c1=for_in_cond(i,n/2,cond,arg_cond,body,arg_body);
 c2=for_in_cond(i+n/2,n-n/2,cond,arg_cond,body,arg_body);
 return c1+c2;
}
//for_reduce runs n iterations
//each iteration produces a scalar value by application of get(i,arg_get)
//the binary tree of iterations reduces the produced values according to the reduce function
//for_reduce returns the reduction value
int for_reduce(int i, int n, int rnv, int (*get)(int, void *), void *arg_get, int (*reduce)(int, int, int, int, void *), void *arg_reduce){
 int g1, g2, f1, f2;
 if (n==1){
  return reduce(i,n,get(i,arg_get),rnv,arg_reduce);
 }
 if (n==2){
  g1=get(i,arg_get); g2=get(i+1,arg_get);
  return reduce(i,n,g1,g2,arg_reduce);
 }
 f1=for_reduce(i, n/2, rnv, get, arg_get, reduce, arg_reduce);
 f2=for_reduce(i+n/2, n-n/2, rnv, get, arg_get, reduce, arg_reduce);
 return reduce(i,n,f1,f2,arg_reduce);
}
//for_ex_cond_reduce runs n iterations
//each iteration such that cond(i,arg_cond) is false produces a scalar value by application of get(i,arg_get) 
//cond is an excluding condition : iteration i produces the neutral value rnv when cond(i,arg_cond) is true
//the binary tree of iterations reduces the produced values according to the reduce function
//the reduction value is in *r
//for_cond_reduce returns the number of non excluded iterations (from 0 to n)
int for_ex_cond_reduce(int i, int n, int rnv, int (*cond)(int, void *), void *arg_cond, int (*get)(int, void *), void *arg_get, int (*reduce)(int, int, int, int, void *), void *arg_reduce){
 int c1, c2, r, r1, r2;
 if (n==1){
  c1=cond(i,arg_cond);
  if (!c1) r=reduce(i,n,get(i,arg_get),rnv,arg_reduce);
  else r=reduce(i,n,rnv,rnv,arg_reduce);
  return r;
 }
 if (n==2){
  c1=cond(i,arg_cond);
  if (!c1) r1=get(i,arg_get);
  else r1=rnv;
  c2=cond(i+1,arg_cond);
  if (!c2) r2=get(i+1,arg_get);
  else r2=rnv;
  r=reduce(i,n,r1,r2,arg_reduce);
  return r;
 }
 r1=for_ex_cond_reduce(i,n/2,rnv,cond,arg_cond,get,arg_get,reduce,arg_reduce);
 r2=for_ex_cond_reduce(i+n/2,n-n/2,rnv,cond,arg_cond,get,arg_get,reduce,arg_reduce);
 r=reduce(i,n,r1,r2,arg_reduce);
 return r;
}
//for_in_cond_reduce runs n iterations
//each iteration such that cond(i,arg_cond) is true produces a scalar value by application of get(i,arg_get) 
//cond is an including condition : iteration i produces the neutral value rnv when cond(i,arg_cond) is false
//the binary tree of iterations reduces the produced values according to the reduce function
//the reduction value is in *r
//for_in_cond_reduce returns the number of non excluded iterations (from 0 to n)
int for_in_cond_reduce(int i, int n, int rnv, int (*cond)(int, void *), void *arg_cond, int (*get)(int, void *), void *arg_get, int (*reduce)(int, int, int, int, void *), void *arg_reduce){
 int c1, c2, r, r1, r2;
 if (n==1){
  c1=cond(i,arg_cond);
  if (c1) r=reduce(i,n,get(i,arg_get),rnv,arg_reduce);
  else r=reduce(i,n,rnv,rnv,arg_reduce);
  return r;
 }
 if (n==2){
  c1=cond(i,arg_cond);
  if (c1) r1=get(i,arg_get);
  else r1=rnv;
  c2=cond(i+1,arg_cond);
  if (c2) r2=get(i+1,arg_get);
  else r2=rnv;
  r=reduce(i,n,r1,r2,arg_reduce);
  return r;
 }
 r1=for_in_cond_reduce(i,n/2,rnv,cond,arg_cond,get,arg_get,reduce,arg_reduce);
 r2=for_in_cond_reduce(i+n/2,n-n/2,rnv,cond,arg_cond,get,arg_get,reduce,arg_reduce);
 r=reduce(i,n,r1,r2,arg_reduce);
 return r;
}
//the reduction can be a sum (reduce(i,j,a,b,arg_reduce)=a+b and rnv=0)
//the reduction can be a product (reduce(i,j,a,b,arg_reduce)=a*b and rnv=1)
//the reduction can be a max (reduce(i,j,a,b,arg_reduce)=(a>b)?a:b and rnv=-1)
//the reduction can be a min (reduce(i,j,a,b,arg_reduce)=(a<b)?a:b and rnv=MAXINT)
//the reduction can be the index of the max (reduce(i,j,a,b,arg_reduce)=(a>b)?i:j and rnv=-1)
//the reduction can be the index of the min (reduce(i,j,a,b,arg_reduce)=(a<b)?i:j and rnv=MAXINDEX)
//the reduction can be the greatest index (reduce(i,j,a,b,arg_reduce)=(i>j)?i:j and rnv=-1)
//the reduction can be the smallest index (reduce(i,j,a,b,arg_reduce)=(i<j)?i:j and rnv=MAXINDEX)
int reduce_sum(int i, int n, int a, int b, void *arg){
 return a+b;
}
int reduce_product(int i, int n, int a, int b, void *arg){
 return a*b;
}
int reduce_max(int i, int n, int a, int b, void *arg){
 return (a>b)?a:b;
}
int reduce_min(int i, int n, int a, int b, void *arg){
 return (a<b)?a:b;
}
int reduce_index_of_max(int i, int n, int a, int b, void *arg){
 return (a>b)?i:i+n/2;
}
int reduce_index_of_min(int i, int n, int a, int b, void *arg){
 return (a<b)?i:i+n/2;
}
int reduce_gt_index(int i, int n, int a, int b, void *arg){
 int j=i+n/2;
 return (i>j)?i:j;
}
int reduce_lt_index(int i, int n, int a, int b, void *arg){
 int j=i+n/2;
 return (i<j)?i:j;
}
void produce_int(int k, int i){
 int j;
 for (j=0;j<i;j++) send_int(k, receive_int(k));
}
void consume_int(int k, int i, int size){
 int j;
 for (j=0;j<size-i;j++) send_int(k, receive_int(k));
}
void update_int(int k, int size){
 int j;
 for (j=0;j<size;j++) send_int(k, receive_int(k));
}
void for_loop_produce_int(int k, int i, int n,
           void (*body)(int, int, void *), void *arg_body, void (*produce_int)(int, int)){
 if (n==1){body(k, i, arg_body); produce_int(k, i+1); return;}
 if (n==2){
  body(k, i, arg_body); body(k, i+1, arg_body); produce_int(k, i+2);
  return;
 }
 for_loop_produce_int(k, i, n/2, body, arg_body,produce_int);
 for_loop_produce_int(k, i+n/2, n-n/2, body, arg_body,produce_int);
}
void for_loop_consume_int(int k, int i, int n, int size,
           void (*body)(int, int, int, void *), void *arg_body, void (*consume_int)(int, int, int)){
 if (n==1){body(k, i, size, arg_body); consume_int(k, i+1, size); return;}
 if (n==2){
  body(k, i, size, arg_body); body(k, i+1, size, arg_body); consume_int(k, i+2, size);
  return;
 }
 for_loop_consume_int(k, i, n/2, size, body, arg_body,consume_int);
 for_loop_consume_int(k, i+n/2, n-n/2, size, body, arg_body,consume_int);
}
void for_loop_update_int(int k, int i, int n, int size,
           void (*body)(int, int, int, void *), void *arg_body,
           void (*update_int)(int, int)){
 if (n==1){body(k, i, size, arg_body); update_int(k, size); return;}
 if (n==2){
  body(k, i, size, arg_body); body(k, i+1, size, arg_body); update_int(k, size);
  return;
 }
 for_loop_update_int(k, i, n/2, size, body, arg_body, update_int);
 for_loop_update_int(k, i+n/2, n-n/2, size, body, arg_body, update_int);
}
int canal[16][1024];
int h[16],t[16];
void init_canal(){
 int i;
 for (i=0;i<16;i++){
  h[i]=t[i]=0;
 }
}
void send_int(int k, int v){
 canal[k][t[k]]=v;
 t[k]=t[k]+1;
 if (t[k]==1024) t[k]=0;
}
int receive_int(int k){
 int x=canal[k][h[k]];
 h[k]=h[k]+1;
 if (h[k]==1024) h[k]=0;
 return x;
}
void dump_canal(int k){
 int i, head, tail;
 printf("head of canal %d at %d\n",k,h[k]);
 printf("tail of canal %d at %d\n",k,t[k]);
 head=h[k]; tail=t[k];
 if (tail<head) tail+=1024;
 for (i=head; i<tail; i++)
  if (i<1024) printf("canal[%d]=%d\n",i-head,canal[k][i]);
  else printf("canal[%d]=%d\n",i-1024-head,canal[k][i-1024]);
}
