#define NULL 0
#include "par_for.h"
//while_loop launches iterations while cond is satisfied (cond is the loop exit condition)
//while_loop assumes that if cond is true for i, it is true for any j>i
//while_loop returns the number of iterations in the loop
//for each iteration in the loop, body(i,arg_body) is applied
//while_loop returns the number of non excluded iterations
int while_loop(int i, int n, int (*cond)(int, void *), void *arg_cond, void (*body)(int, void *), void *arg_body){
 int nb_iter=for_ex_cond(i,n,cond,arg_cond,body,arg_body);
 if (nb_iter==n)
   nb_iter+=while_loop(i+n,2*n,cond,arg_cond,body,arg_body);
 return nb_iter;
}
//until_loop launches iterations until cond is satisfied (!cond is the loop exit condition)
//until_loop assumes that if !cond is true for i, it is true for any j>i
//until_loop returns the number of iterations in the loop
//for each iteration in the loop, body(i,arg_body) is applied
//until_loop returns the number of non excluded iterations
int until_loop(int i, int n, int (*cond)(int, void *), void *arg_cond, void (*body)(int, void *), void *arg_body){
 int nb_iter=for_in_cond(i,n,cond,arg_cond,body,arg_body);
 if (nb_iter==n)
   nb_iter+=until_loop(i+n,2*n,cond,arg_cond,body,arg_body);
 return nb_iter;
}
static void empty(){}
//while_reduce launches iterations until cond is satisfied
//each iteration such that cond(i,arg_cond) is false produces a scalar value by application of get(i,arg_get) 
//cond is an excluding condition : iteration i produces the neutral value rnv when cond(i,arg_cond) is true
//the binary tree of iterations reduces the produced values according to the reduce function
//while_reduce returns the reduction value
int while_reduce(int i, int n, int rnv, int (*cond)(int, void *), void *arg_cond, int (*get)(int, void *), void *arg_get, int (*reduce)(int, int, int, int, void *), void *arg_reduce){
 int r1, r2;
 int nb_iter=for_ex_cond(i,n,cond,arg_cond,empty,NULL);
 r1=for_ex_cond_reduce(i,n,rnv,cond,arg_cond,get,arg_get,reduce,arg_reduce);
 if (nb_iter==n){
   r2 = while_reduce(i+n,2*n,rnv,cond,arg_cond,get,arg_get,reduce,arg_reduce);
   r1=reduce(i,n,r1,r2,arg_reduce);
 }
 return r1; 
}
//until_reduce launches iterations until cond is no more satisfied
//each iteration such that cond(i,arg_cond) is true produces a scalar value by application of get(i,arg_get) 
//cond is an including condition : iteration i produces the neutral value rnv when cond(i,arg_cond) is false
//the binary tree of iterations reduces the produced values according to the reduce function
//until_reduce returns the reduction value
int until_reduce(int i, int n, int rnv, int (*cond)(int, void *), void *arg_cond, int (*get)(int, void *), void *arg_get, int (*reduce)(int, int, int, int, void *), void *arg_reduce){
 int r1, r2;
 int nb_iter=for_in_cond(i,n,cond,arg_cond,empty,NULL);
 r1=for_in_cond_reduce(i,n,rnv,cond,arg_cond,get,arg_get,reduce,arg_reduce);
 if (nb_iter==n){
   r2 = until_reduce(i+n,2*n,rnv,cond,arg_cond,get,arg_get,reduce,arg_reduce);
   r1=reduce(i,n,r1,r2,arg_reduce);
 }
 return r1; 
}
